@@ -1,6 +1,6 @@
 # Copyright (c) 2009-2012 Denis Bilenko. See LICENSE for details.
 # cython: auto_pickle=False,embedsignature=True,always_allow_keywords=False
-
+# pylint:disable=too-many-lines
 from __future__ import absolute_import, print_function, division
 
 from sys import _getframe as sys_getframe
@@ -25,7 +25,6 @@ from gevent._hub_primitives import wait_on_objects as wait
 from gevent.timeout import Timeout
 
 from gevent._config import config as GEVENT_CONFIG
-from gevent._util import Lazy
 from gevent._util import readproperty
 from gevent._hub_local import get_hub_noargs as get_hub
 from gevent import _waiter
@@ -45,7 +44,20 @@ __all__ = [
 locals()['getcurrent'] = __import__('greenlet').getcurrent
 locals()['greenlet_init'] = lambda: None
 locals()['Waiter'] = _waiter.Waiter
+# With Cython, this raises a TypeError if the parent is *not*
+# the hub (SwitchOutGreenletWithLoop); in pure-Python, we will
+# very likely get an AttributeError immediately after when we access `loop`;
+# The TypeError message is more informative on Python 2.
+# This must ONLY be called when we know that `s` is not None and is in fact a greenlet
+# object (e.g., when called on `self`)
+locals()['get_my_hub'] = lambda s: s.parent
+# This must also ONLY be called when we know that S is not None and is in fact a greenlet
+# object (including the result of getcurrent())
+locals()['get_generic_parent'] = lambda s: s.parent
 
+# Frame access
+locals()['get_f_back'] = lambda frame: frame.f_back
+locals()['get_f_lineno'] = lambda frame: frame.f_lineno
 
 if _PYPY:
     import _continuation # pylint:disable=import-error
@@ -113,32 +125,49 @@ class _Frame(object):
 
     __slots__ = ('f_code', 'f_lineno', 'f_back')
 
-    def __init__(self, f_code, f_lineno, f_back):
-        self.f_code = f_code
-        self.f_lineno = f_lineno
-        self.f_back = f_back
+    def __init__(self):
+        self.f_code = None
+        self.f_back = None
+        self.f_lineno = 0
 
     @property
     def f_globals(self):
         return None
 
-def _Frame_from_list(frames):
-    previous = None
-    for frame in reversed(frames):
-        f = _Frame(frame[0], frame[1], previous)
-        previous = f
-    return previous
 
 def _extract_stack(limit):
-    frame = sys_getframe()
-    frames = []
+    try:
+        frame = sys_getframe()
+    except ValueError:
+        # In certain embedded cases that directly use the Python C api
+        # to call Greenlet.spawn (e.g., uwsgi) this can raise
+        # `ValueError: call stack is not deep enough`. This is because
+        # the Cython stack frames for Greenlet.spawn ->
+        # Greenlet.__init__ -> _extract_stack are all on the C level,
+        # not the Python level.
+        # See https://github.com/gevent/gevent/issues/1212
+        frame = None
+
+    newest_Frame = None
+    newer_Frame = None
 
     while limit and frame is not None:
         limit -= 1
-        frames.append((frame.f_code, frame.f_lineno))
-        frame = frame.f_back
+        older_Frame = _Frame()
+        # Arguments are always passed to the constructor as Python objects,
+        # meaning we wind up boxing the f_lineno just to unbox it if we pass it.
+        # It's faster to simply assign once the object is created.
+        older_Frame.f_code = frame.f_code
+        older_Frame.f_lineno = get_f_lineno(frame) # pylint:disable=undefined-variable
+        if newer_Frame is not None:
+            newer_Frame.f_back = older_Frame
+        newer_Frame = older_Frame
+        if newest_Frame is None:
+            newest_Frame = newer_Frame
 
-    return frames
+        frame = get_f_back(frame) # pylint:disable=undefined-variable
+
+    return newest_Frame
 
 
 _greenlet__init__ = greenlet.__init__
@@ -169,6 +198,10 @@ class Greenlet(greenlet):
            a false value to disable ``spawn_tree_locals``, ``spawning_greenlet``,
            and ``spawning_stack``. The first two will be None in that case, and the
            latter will be empty.
+
+        .. versionchanged:: 1.5
+           Greenlet objects are now more careful to verify that their ``parent`` is really
+           a gevent hub, raising a ``TypeError`` earlier instead of an ``AttributeError`` later.
         """
         # The attributes are documented in the .rst file
 
@@ -198,6 +231,15 @@ class Greenlet(greenlet):
         # 2.7.14       : Mean +- std dev: 3.37 us +- 0.20 us
         # PyPy2 5.10.0 : Mean +- std dev: 4.44 us +- 0.28 us
 
+        # Switching to reified frames and some more tuning gets us here:
+        # 3.7.2        : Mean +- std dev: 2.53 us +- 0.15 us
+        # 2.7.16       : Mean +- std dev: 2.35 us +- 0.12 us
+        # PyPy2 7.1    : Mean +- std dev: 11.6 us +- 0.4 us
+
+        # Compared to the released 1.4 (tested at the same time):
+        # 3.7.2        : Mean +- std dev: 3.21 us +- 0.32 us
+        # 2.7.16       : Mean +- std dev: 3.11 us +- 0.19 us
+        # PyPy2 7.1    : Mean +- std dev: 12.3 us +- 0.8 us
 
         _greenlet__init__(self, None, get_hub())
 
@@ -218,8 +260,10 @@ class Greenlet(greenlet):
         #: start() and start_later() as those two objects, respectively.
         #: Once this becomes non-None, the Greenlet cannot be started again. Conversely,
         #: kill() and throw() check for non-None to determine if this object has ever been
-        #: scheduled for starting. A placeholder _dummy_event is assigned by them to prevent
+        #: scheduled for starting. A placeholder _cancelled_start_event is assigned by them to prevent
         #: the greenlet from being started in the future, if necessary.
+        #: In the usual case, this transitions as follows: None -> event -> _start_completed_event.
+        #: A value of None means we've never been started.
         self._start_event = None
 
         self._notifier = None
@@ -239,36 +283,44 @@ class Greenlet(greenlet):
                 self.spawn_tree_locals = spawner.spawn_tree_locals
             except AttributeError:
                 self.spawn_tree_locals = {}
-                if spawner.parent is not None:
+                if get_generic_parent(spawner) is not None: # pylint:disable=undefined-variable
                     # The main greenlet has no parent.
                     # Its children get separate locals.
                     spawner.spawn_tree_locals = self.spawn_tree_locals
 
-            self._spawning_stack_frames = _extract_stack(self.spawning_stack_limit)
-            self._spawning_stack_frames.extend(getattr(spawner, '_spawning_stack_frames', []))
+            self.spawning_stack = _extract_stack(self.spawning_stack_limit)
+            # Don't copy the spawning greenlet's
+            # '_spawning_stack_frames' into ours. That's somewhat
+            # confusing, and, if we're not careful, a deep spawn tree
+            # can lead to excessive memory usage (an infinite spawning
+            # tree could lead to unbounded memory usage without care
+            # --- see https://github.com/gevent/gevent/issues/1371)
+            # The _spawning_stack_frames may be cleared out later if we access spawning_stack
         else:
             # None is the default for all of these in Cython, but we
             # need to declare them for pure-Python mode.
             self.spawning_greenlet = None
             self.spawn_tree_locals = None
-            self._spawning_stack_frames = None
-
-    @Lazy
-    def spawning_stack(self):
-        # Store this in the __dict__. We don't use it from the C
-        # code. It's tempting to discard _spawning_stack_frames
-        # after this, but child greenlets may still be created
-        # that need it.
-        return _Frame_from_list(self._spawning_stack_frames or [])
+            self.spawning_stack = None
 
     def _get_minimal_ident(self):
-        reg = self.parent.ident_registry
+        # Helper function for cython, to allow typing `reg` and making a
+        # C call to get_ident.
+
+        # If we're being accessed from a hub different than the one running
+        # us, aka get_hub() is not self.parent, then calling hub.ident_registry.get_ident()
+        # may be quietly broken: it's not thread safe.
+        # If our parent is no longer the hub for whatever reason, this will raise a
+        # AttributeError or TypeError.
+        hub = get_my_hub(self) # pylint:disable=undefined-variable
+
+        reg = hub.ident_registry
         return reg.get_ident(self)
 
     @property
     def minimal_ident(self):
         """
-        A small, unique integer that identifies this object.
+        A small, unique non-negative integer that identifies this object.
 
         This is similar to :attr:`threading.Thread.ident` (and `id`)
         in that as long as this object is alive, no other greenlet *in
@@ -278,10 +330,16 @@ class Greenlet(greenlet):
         will be available for reuse.
 
         To get ids that are unique across all hubs, combine this with
-        the hub's ``minimal_ident``.
+        the hub's (``self.parent``) ``minimal_ident``.
+
+        Accessing this property from threads other than the thread running
+        this greenlet is not defined.
 
         .. versionadded:: 1.3a2
+
         """
+        # Not @Lazy, implemented manually because _ident is in the structure
+        # of the greenlet for fast access
         if self._ident is None:
             self._ident = self._get_minimal_ident()
         return self._ident
@@ -291,11 +349,16 @@ class Greenlet(greenlet):
         """
         The greenlet name. By default, a unique name is constructed using
         the :attr:`minimal_ident`. You can assign a string to this
-        value to change it. It is shown in the `repr` of this object.
+        value to change it. It is shown in the `repr` of this object if it
+        has been assigned to or if the `minimal_ident` has already been generated.
 
         .. versionadded:: 1.3a2
+        .. versionchanged:: 1.4
+           Stop showing generated names in the `repr` when the ``minimal_ident``
+           hasn't been requested. This reduces overhead and may be less confusing,
+           since ``minimal_ident`` can get reused.
         """
-        return 'Greenlet-%d' % (self.minimal_ident)
+        return 'Greenlet-%d' % (self.minimal_ident,)
 
     def _raise_exception(self):
         reraise(*self.exc_info)
@@ -303,7 +366,8 @@ class Greenlet(greenlet):
     @property
     def loop(self):
         # needed by killall
-        return self.parent.loop
+        hub = get_my_hub(self) # type:SwitchOutGreenletWithLoop pylint:disable=undefined-variable
+        return hub.loop
 
     def __nonzero__(self):
         return self._start_event is not None and self._exc_info is None
@@ -321,6 +385,7 @@ class Greenlet(greenlet):
         @property
         def dead(self):
             "Boolean indicating that the greenlet is dead and will not run again."
+            # pylint:disable=no-member
             if self._greenlet__main:
                 return False
             if self.__start_cancelled_by_kill() or self.__started_but_aborted():
@@ -330,15 +395,31 @@ class Greenlet(greenlet):
     else:
         @property
         def dead(self):
-            "Boolean indicating that the greenlet is dead and will not run again."
-            return self.__start_cancelled_by_kill() or self.__started_but_aborted() or greenlet.dead.__get__(self)
+            """
+            Boolean indicating that the greenlet is dead and will not run again.
+
+            This is true if:
+
+            1. We were never started, but were :meth:`killed <kill>`
+               immediately after creation (not possible with :meth:`spawn`); OR
+            2. We were started, but were killed before running; OR
+            3. We have run and terminated (by raising an exception out of the
+               started function or by reaching the end of the started function).
+            """
+            return (
+                self.__start_cancelled_by_kill()
+                or self.__started_but_aborted()
+                or greenlet.dead.__get__(self)
+            )
 
     def __never_started_or_killed(self):
         return self._start_event is None
 
     def __start_pending(self):
-        return (self._start_event is not None
-                and (self._start_event.pending or getattr(self._start_event, 'active', False)))
+        return (
+            self._start_event is not None
+            and (self._start_event.pending or getattr(self._start_event, 'active', False))
+        )
 
     def __start_cancelled_by_kill(self):
         return self._start_event is _cancelled_start_event
@@ -347,10 +428,12 @@ class Greenlet(greenlet):
         return self._start_event is _start_completed_event
 
     def __started_but_aborted(self):
-        return (not self.__never_started_or_killed() # we have been started or killed
-                and not self.__start_cancelled_by_kill() # we weren't killed, so we must have been started
-                and not self.__start_completed() # the start never completed
-                and not self.__start_pending()) # and we're not pending, so we must have been aborted
+        return (
+            not self.__never_started_or_killed() # we have been started or killed
+            and not self.__start_cancelled_by_kill() # we weren't killed, so we must have been started
+            and not self.__start_completed() # the start never completed
+            and not self.__start_pending() # and we're not pending, so we must have been aborted
+        )
 
     def __cancel_start(self):
         if self._start_event is None:
@@ -366,22 +449,37 @@ class Greenlet(greenlet):
         self._start_event.close()
 
     def __handle_death_before_start(self, args):
-        # args is (t, v, tb) or simply t or v
+        # args is (t, v, tb) or simply t or v.
+        # The last two cases are transformed into (t, v, None);
+        # if the single argument is an exception type, a new instance
+        # is created; if the single argument is not an exception type and also
+        # not an exception, it is wrapped in a BaseException (this is not
+        # documented, but should result in better behaviour in the event of a
+        # user error---instead of silently printing something to stderr, we still
+        # kill the greenlet).
         if self._exc_info is None and self.dead:
-            # the greenlet was never switched to before and it will never be, _report_error was not called
-            # the result was not set and the links weren't notified. let's do it here.
-            # checking that self.dead is true is essential, because throw() does not necessarily kill the greenlet
-            # (if the exception raised by throw() is caught somewhere inside the greenlet).
+            # the greenlet was never switched to before and it will
+            # never be; _report_error was not called, the result was
+            # not set, and the links weren't notified. Let's do it
+            # here.
+            #
+            # checking that self.dead is true is essential, because
+            # throw() does not necessarily kill the greenlet (if the
+            # exception raised by throw() is caught somewhere inside
+            # the greenlet).
             if len(args) == 1:
                 arg = args[0]
-                #if isinstance(arg, type):
-                if type(arg) is type(Exception):
+                if isinstance(arg, type) and issubclass(arg, BaseException):
                     args = (arg, arg(), None)
                 else:
                     args = (type(arg), arg, None)
             elif not args:
                 args = (GreenletExit, GreenletExit(), None)
-            self._report_error(args)
+            if not issubclass(args[0], BaseException):
+                # Random non-type, non-exception arguments.
+                args = (BaseException, BaseException(args), None)
+            assert issubclass(args[0], BaseException)
+            self.__report_error(args)
 
     @property
     def started(self):
@@ -415,7 +513,14 @@ class Greenlet(greenlet):
 
     def __repr__(self):
         classname = self.__class__.__name__
-        result = '<%s "%s" at %s' % (classname, self.name, hex(id(self)))
+        # If no name has been assigned, don't generate one, including a minimal_ident,
+        # if not necessary. This reduces the use of weak references and associated
+        # overhead.
+        if 'name' not in self.__dict__ and self._ident is None:
+            name = ' '
+        else:
+            name = ' "%s" ' % (self.name,)
+        result = '<%s%sat %s' % (classname, name, hex(id(self)))
         formatted = self._formatinfo()
         if formatted:
             result += ': ' + formatted
@@ -470,7 +575,12 @@ class Greenlet(greenlet):
         """
         ei = self._exc_info
         if ei is not None and ei[0] is not None:
-            return (ei[0], ei[1], load_traceback(ei[2]))
+            return (
+                ei[0],
+                ei[1],
+                # The pickled traceback may be None if we couldn't pickle it.
+                load_traceback(ei[2]) if ei[2] else None
+            )
 
     def throw(self, *args):
         """Immediately switch into the greenlet and raise an exception in it.
@@ -497,7 +607,9 @@ class Greenlet(greenlet):
     def start(self):
         """Schedule the greenlet to run in this loop iteration"""
         if self._start_event is None:
-            self._start_event = self.parent.loop.run_callback(self.switch)
+            _call_spawn_callbacks(self)
+            hub = get_my_hub(self) # type:SwitchOutGreenletWithLoop pylint:disable=undefined-variable
+            self._start_event = hub.loop.run_callback(self.switch)
 
     def start_later(self, seconds):
         """
@@ -507,9 +619,47 @@ class Greenlet(greenlet):
         *seconds* later
         """
         if self._start_event is None:
-            self._start_event = self.parent.loop.timer(seconds)
+            _call_spawn_callbacks(self)
+            hub = get_my_hub(self) # pylint:disable=undefined-variable
+            self._start_event = hub.loop.timer(seconds)
             self._start_event.start(self.switch)
 
+    @staticmethod
+    def add_spawn_callback(callback):
+        """
+        add_spawn_callback(callback) -> None
+
+        Set up a *callback* to be invoked when :class:`Greenlet` objects
+        are started.
+
+        The invocation order of spawn callbacks is unspecified.  Adding the
+        same callback more than one time will not cause it to be called more
+        than once.
+
+        .. versionadded:: 1.4.0
+        """
+        global _spawn_callbacks
+        if _spawn_callbacks is None:  # pylint:disable=used-before-assignment
+            _spawn_callbacks = set()
+        _spawn_callbacks.add(callback)
+
+    @staticmethod
+    def remove_spawn_callback(callback):
+        """
+        remove_spawn_callback(callback) -> None
+
+        Remove *callback* function added with :meth:`Greenlet.add_spawn_callback`.
+        This function will not fail if *callback* has been already removed or
+        if *callback* was never added.
+
+        .. versionadded:: 1.4.0
+        """
+        global _spawn_callbacks
+        if _spawn_callbacks is not None:
+            _spawn_callbacks.discard(callback)
+            if not _spawn_callbacks:
+                _spawn_callbacks = None
+
     @classmethod
     def spawn(cls, *args, **kwargs):
         """
@@ -551,14 +701,27 @@ class Greenlet(greenlet):
         g.start_later(seconds)
         return g
 
+    def _maybe_kill_before_start(self, exception):
+        # Helper for Greenlet.kill(), and also for killall()
+        self.__cancel_start()
+        self.__free()
+        dead = self.dead
+        if dead:
+            self.__handle_death_before_start((exception,))
+        return dead
+
     def kill(self, exception=GreenletExit, block=True, timeout=None):
         """
         Raise the ``exception`` in the greenlet.
 
-        If ``block`` is ``True`` (the default), wait until the greenlet dies or the optional timeout expires.
+        If ``block`` is ``True`` (the default), wait until the greenlet
+        dies or the optional timeout expires; this may require switching
+        greenlets.
         If block is ``False``, the current greenlet is not unscheduled.
 
-        The function always returns ``None`` and never raises an error.
+        This function always returns ``None`` and never raises an error. It
+        may be called multpile times on the same greenlet object, and may be
+        called on an unstarted or dead greenlet.
 
         .. note::
 
@@ -577,9 +740,11 @@ class Greenlet(greenlet):
 
             Use care when killing greenlets. If the code executing is not
             exception safe (e.g., makes proper use of ``finally``) then an
-            unexpected exception could result in corrupted state.
+            unexpected exception could result in corrupted state. Using
+            a :meth:`link` or :meth:`rawlink` (cheaper) may be a safer way to
+            clean up resources.
 
-        See also :func:`gevent.kill`.
+        See also :func:`gevent.kill` and :func:`gevent.killall`.
 
         :keyword type exception: The type of exception to raise in the greenlet. The default
             is :class:`GreenletExit`, which indicates a :meth:`successful` completion
@@ -588,20 +753,17 @@ class Greenlet(greenlet):
         .. versionchanged:: 0.13.0
             *block* is now ``True`` by default.
         .. versionchanged:: 1.1a2
-            If this greenlet had never been switched to, killing it will prevent it from ever being switched to.
+            If this greenlet had never been switched to, killing it will
+            prevent it from *ever* being switched to. Links (:meth:`rawlink`)
+            will still be executed, though.
         """
-        self.__cancel_start()
-
-        if self.dead:
-            self.__handle_death_before_start((exception,))
-        else:
+        if not self._maybe_kill_before_start(exception):
             waiter = Waiter() if block else None # pylint:disable=undefined-variable
-            self.parent.loop.run_callback(_kill, self, exception, waiter)
-            if block:
+            hub = get_my_hub(self) # pylint:disable=undefined-variable
+            hub.loop.run_callback(_kill, self, exception, waiter)
+            if waiter is not None:
                 waiter.get()
                 self.join(timeout)
-        # it should be OK to use kill() in finally or kill a greenlet from more than one place;
-        # thus it should not raise when the greenlet is already killed (= not started)
 
     def get(self, block=True, timeout=None):
         """
@@ -628,7 +790,7 @@ class Greenlet(greenlet):
         try:
             t = Timeout._start_new_or_dummy(timeout)
             try:
-                result = self.parent.switch()
+                result = get_my_hub(self).switch() # pylint:disable=undefined-variable
                 if result is not self:
                     raise InvalidSwitchError('Invalid switch into Greenlet.get(): %r' % (result, ))
             finally:
@@ -662,7 +824,7 @@ class Greenlet(greenlet):
         try:
             t = Timeout._start_new_or_dummy(timeout)
             try:
-                result = self.parent.switch()
+                result = get_my_hub(self).switch() # pylint:disable=undefined-variable
                 if result is not self:
                     raise InvalidSwitchError('Invalid switch into Greenlet.join(): %r' % (result, ))
             finally:
@@ -675,24 +837,32 @@ class Greenlet(greenlet):
             self.unlink(switch)
             raise
 
-    def _report_result(self, result):
+    def __report_result(self, result):
         self._exc_info = (None, None, None)
         self.value = result
         if self._links and not self._notifier:
-            self._notifier = self.parent.loop.run_callback(self._notify_links)
+            hub = get_my_hub(self) # pylint:disable=undefined-variable
+            self._notifier = hub.loop.run_callback(self._notify_links)
 
-    def _report_error(self, exc_info):
+    def __report_error(self, exc_info):
         if isinstance(exc_info[1], GreenletExit):
-            self._report_result(exc_info[1])
+            self.__report_result(exc_info[1])
             return
 
-        self._exc_info = exc_info[0], exc_info[1], dump_traceback(exc_info[2])
+        # Depending on the error, we may not be able to pickle it.
+        # In particular, RecursionError can be a problem.
+        try:
+            tb = dump_traceback(exc_info[2])
+        except: # pylint:disable=bare-except
+            tb = None
+        self._exc_info = exc_info[0], exc_info[1], tb
 
+        hub = get_my_hub(self) # pylint:disable=undefined-variable
         if self._links and not self._notifier:
-            self._notifier = self.parent.loop.run_callback(self._notify_links)
+            self._notifier = hub.loop.run_callback(self._notify_links)
 
         try:
-            self.parent.handle_error(self, *exc_info)
+            hub.handle_error(self, *exc_info)
         finally:
             del exc_info
 
@@ -704,13 +874,23 @@ class Greenlet(greenlet):
             try:
                 result = self._run(*self.args, **self.kwargs)
             except: # pylint:disable=bare-except
-                self._report_error(sys_exc_info())
-                return
-            self._report_result(result)
+                self.__report_error(sys_exc_info())
+            else:
+                self.__report_result(result)
         finally:
-            self.__dict__.pop('_run', None)
-            self.args = ()
-            self.kwargs.clear()
+            self.__free()
+
+    def __free(self):
+        try:
+            # It seems that Cython 0.29.13 sometimes miscompiles
+            # self.__dict__.pop('_run', None) ? When we moved this out of the
+            # inline finally: block in run(), we started getting strange
+            # exceptions from places that subclassed Greenlet.
+            del self._run
+        except AttributeError:
+            pass
+        self.args = ()
+        self.kwargs.clear()
 
     def _run(self):
         """
@@ -737,13 +917,16 @@ class Greenlet(greenlet):
         The *callback* will be called with this instance as an
         argument.
 
-        .. caution:: The callable will be called in the HUB greenlet.
+        .. caution::
+            The *callback* will be called in the hub and
+            **MUST NOT** raise an exception.
         """
         if not callable(callback):
             raise TypeError('Expected callable: %r' % (callback, ))
         self._links.append(callback) # pylint:disable=no-member
         if self.ready() and self._links and not self._notifier:
-            self._notifier = self.parent.loop.run_callback(self._notify_links)
+            hub = get_my_hub(self) # pylint:disable=undefined-variable
+            self._notifier = hub.loop.run_callback(self._notify_links)
 
     def link(self, callback, SpawnedLink=SpawnedLink):
         """
@@ -802,8 +985,8 @@ class Greenlet(greenlet):
             link = self._links.pop(0)
             try:
                 link(self)
-            except: # pylint:disable=bare-except
-                self.parent.handle_error((link, self), *sys_exc_info())
+            except: # pylint:disable=bare-except, undefined-variable
+                get_my_hub(self).handle_error((link, self), *sys_exc_info())
 
 
 class _dummy_event(object):
@@ -825,12 +1008,14 @@ _cancelled_start_event = _dummy_event()
 _start_completed_event = _dummy_event()
 
 
+# This is *only* called as a callback from the hub via Greenlet.kill(),
+# and its first argument is the Greenlet. So we can be sure about the types.
 def _kill(glet, exception, waiter):
     try:
         glet.throw(exception)
-    except: # pylint:disable=bare-except
+    except: # pylint:disable=bare-except, undefined-variable
         # XXX do we need this here?
-        glet.parent.handle_error(glet, *sys_exc_info())
+        get_my_hub(glet).handle_error(glet, *sys_exc_info())
     if waiter is not None:
         waiter.switch(None)
 
@@ -864,8 +1049,8 @@ def _killall3(greenlets, exception, waiter):
         if not g.dead:
             try:
                 g.throw(exception)
-            except: # pylint:disable=bare-except
-                g.parent.handle_error(g, *sys_exc_info())
+            except: # pylint:disable=bare-except, undefined-variable
+                get_my_hub(g).handle_error(g, *sys_exc_info())
             if not g.dead:
                 diehards.append(g)
     waiter.switch(diehards)
@@ -876,20 +1061,31 @@ def _killall(greenlets, exception):
         if not g.dead:
             try:
                 g.throw(exception)
-            except: # pylint:disable=bare-except
-                g.parent.handle_error(g, *sys_exc_info())
+            except: # pylint:disable=bare-except, undefined-variable
+                get_my_hub(g).handle_error(g, *sys_exc_info())
+
+
+def _call_spawn_callbacks(gr):
+    if _spawn_callbacks is not None:
+        for cb in _spawn_callbacks:
+            cb(gr)
+
+
+_spawn_callbacks = None
 
 
 def killall(greenlets, exception=GreenletExit, block=True, timeout=None):
     """
-    Forceably terminate all the ``greenlets`` by causing them to raise ``exception``.
+    Forceably terminate all the *greenlets* by causing them to raise *exception*.
 
     .. caution:: Use care when killing greenlets. If they are not prepared for exceptions,
        this could result in corrupted state.
 
     :param greenlets: A **bounded** iterable of the non-None greenlets to terminate.
-       *All* the items in this iterable must be greenlets that belong to the same thread.
-    :keyword exception: The exception to raise in the greenlets. By default this is
+       *All* the items in this iterable must be greenlets that belong to the same hub,
+       which should be the hub for this current thread. If this is a generator or iterator
+       that switches greenlets, the results are undefined.
+    :keyword exception: The type of exception to raise in the greenlets. By default this is
         :class:`GreenletExit`.
     :keyword bool block: If True (the default) then this function only returns when all the
         greenlets are dead; the current greenlet is unscheduled during that process.
@@ -905,15 +1101,38 @@ def killall(greenlets, exception=GreenletExit, block=True, timeout=None):
     .. versionchanged:: 1.1a2
         *greenlets* can be any iterable of greenlets, like an iterator or a set.
         Previously it had to be a list or tuple.
+    .. versionchanged:: 1.5a3
+        Any :class:`Greenlet` in the *greenlets* list that hadn't been switched to before
+        calling this method will never be switched to. This makes this function
+        behave like :meth:`Greenlet.kill`. This does not apply to raw greenlets.
+    .. versionchanged:: 1.5a3
+        Now accepts raw greenlets created by :func:`gevent.spawn_raw`.
     """
-    # support non-indexable containers like iterators or set objects
-    greenlets = list(greenlets)
-    if not greenlets:
+
+    need_killed = [] # type: list
+    for glet in greenlets:
+        # Quick pass through to prevent any greenlet from
+        # actually being switched to if it hasn't already.
+        # (Previously we called ``list(greenlets)`` so we're still
+        # linear.)
+        #
+        # We don't use glet.kill() here because we don't want to schedule
+        # any callbacks in the loop; we're about to handle that more directly.
+        try:
+            cancel = glet._maybe_kill_before_start
+        except AttributeError:
+            need_killed.append(glet)
+        else:
+            if not cancel(exception):
+                need_killed.append(glet)
+
+    if not need_killed:
         return
-    loop = greenlets[0].loop
+
+    loop = glet.loop # pylint:disable=undefined-loop-variable
     if block:
         waiter = Waiter() # pylint:disable=undefined-variable
-        loop.run_callback(_killall3, greenlets, exception, waiter)
+        loop.run_callback(_killall3, need_killed, exception, waiter)
         t = Timeout._start_new_or_dummy(timeout)
         try:
             alive = waiter.get()
@@ -922,7 +1141,7 @@ def killall(greenlets, exception=GreenletExit, block=True, timeout=None):
         finally:
             t.cancel()
     else:
-        loop.run_callback(_killall, greenlets, exception)
+        loop.run_callback(_killall, need_killed, exception)
 
 def _init():
     greenlet_init() # pylint:disable=undefined-variable
