@@ -436,14 +436,25 @@ class local(object):
                 return type(type_attr).__get__(type_attr, self, self._local_type)
             # Last case is a non-data descriptor. Instance wins.
             return dct[name]
-        elif name in self._local_type_vars:
-            type_attr = getattr(self._local_type, name)
 
-            # It's not in the dict at all. Is it in the type?
+        if name in self._local_type_vars:
+            # Not in the dictionary, but is found in the type. It could be
+            # a non-data descriptor still. Some descriptors, like @staticmethod,
+            # return objects (functions, in this case), that are *themselves*
+            # descriptors, which when invoked, again, would do the wrong thing.
+            # So we can't rely on getattr() on the type for them, we have to
+            # look through the MRO dicts ourself.
             if name not in self._local_type_get_descriptors:
-                # Not a descriptor, can't execute code
-                return type_attr
-            return type(type_attr).__get__(type_attr, self, self._local_type)
+                # Not a descriptor, can't execute code. So all we need is
+                # the return value of getattr() on our type.
+                return getattr(self._local_type, name)
+
+            for base in self._local_type.mro():
+                bd = base.__dict__
+                if name in bd:
+                    attr_on_type = bd[name]
+                    result = type(attr_on_type).__get__(attr_on_type, self, self._local_type)
+                    return result
 
         # It wasn't in the dict and it wasn't in the type.
         # So the next step is to invoke type(self)__getattr__, if it
@@ -571,7 +582,7 @@ def __new__(cls, *args, **kw):
 try:
     # PyPy2/3 and CPython handle adding a __new__ to the class
     # in different ways. In CPython and PyPy3, it must be wrapped with classmethod;
-    # in PyPy2, it must not. In either case, the args that get passed to
+    # in PyPy2 < 7.3.3, it must not. In either case, the args that get passed to
     # it are stil wrong.
     local.__new__ = 'None'
 except TypeError: # pragma: no cover
@@ -581,7 +592,13 @@ else:
     from gevent._compat import PYPY
     from gevent._compat import PY2
     if PYPY and PY2:
+        # The behaviour changed with no warning between PyPy2 7.3.2 and 7.3.3.
         local.__new__ = __new__
+        try:
+            local() # <= 7.3.2
+        except TypeError:
+            # >= 7.3.3
+            local.__new__ = classmethod(__new__)
     else:
         local.__new__ = classmethod(__new__)
 
